"""

Miscellaneous functions written by Martin Preishuber

"""

from string import strip, lower, zfill, find, atoi, join
from os import getcwd, chdir, mkdir, lstat, unlink, rmdir, utime, wait
from urlparse import urlparse
from time import time
from popen2 import popen3

import signal
import os
import string
import copy
import sys

# Boolean constants

TRUE = 1
FALSE = 0

# Color constants

BLACK = "30"
RED = "31"
GREEN = "32"
YELLOW = "33"
BLUE = "34"
PURPLE = "35"
AQUA = "36"
WHITE = "37"


"""

String operations

"""

# Extended split (doesn't split within special characters)
def split(s, sep = None, special = None):
    if (special == None):
        if (sep == None):
            return string.split(s)
        else:
            return string.split(s, sep)
    else:
        list = []
        elem = ""
        mode = 0
        for i in range(len(s)):
            if (s[i] == sep):
                if (mode == 0) and (elem != ""):
                    list.append(elem)
                    elem = ""
                elif (mode == 1):
                    elem = "%s%s" % (elem, s[i])
            elif (s[i] == special):
                if (mode == 0):
                    mode = 1
                elif (mode == 1):
                    mode = 0
            else:
                elem = "%s%s" % (elem, s[i])
        list.append(elem)
        return list

# Fill a string with a character up to a given width
def cfill(source, char, width):
    while len(source) < width:
        source = "%s%s" % (source, char)
    return source

# Pad a string with a character up to a given width
def pfill(source, char, width):
    while len(source) < width:
        source = char + source
    return source

# Strip blanks (NOT whitespaces) on the left side of a string
def lstrip(string):
    while (string[0] == " ") and (len(string) > 0):
        string = string[1:]
    return string

# Merge two strings byte-by-byte
def merge(string1, string2):
    pos1 = 0
    pos2 = 0
    result = ""
    while (len(string1) >= (pos1 + 1)) and (len(string2) >= (pos2 + 1)):
        result = "%s%s%s" % (result, string1[pos1], string2[pos2])
        pos1 = pos1 + 1
        pos2 = pos2 + 1
    if (len(string1) >= (pos1 + 1)):
        result = "%s%s" % (result, string1[pos1:])
    if (len(string2) >= (pos2 + 1)):
        result = "%s%s" % (result, string2[pos2:])
    return result


"""

RE module functions

"""

def unfold(regexp):
    splitted = split(regexp, "\n")
    for i in range(len(splitted)):
        splitted[i] = strip(splitted[i])
    return join(splitted, "")

"""

Type conversion functions

"""

# Converts a Boolean value to the corresponding String
def bool2str(value):
    if value == FALSE:
        return "False"
    elif value == TRUE:
        return "True"

# Converts a String value to the corresponding Boolean
def str2bool(value):
    if (lower(value) == "true") or (lower(value) == "yes") or (value == "1"):
        return TRUE
    elif (lower(value) == "false") or (lower(value) == "no") or (value == "0"):
        return FALSE
    else:
        return None

# Returns a string representation of a number (with a pretty-print option)
def lng2str(number, prettyprint = TRUE):
    if (prettyprint == TRUE):
        number = list(str(number))
        if number[-1] == "L":
            number = number[:-1]
        number.reverse()
        dots = divmod(len(number) - 1, 3)[0]
        for i in range(dots):
            number.insert(3 + (i * 3) + i, ".")
        number.reverse()
        return join(number, "")
    else:
        return str(number)


"""

List operations

"""

# Strip each value of a list
def striplist(list):
    for i in range(len(list)):
        list[i] = strip(list[i])
    return list

# Removes empty entries from a list
def cleanlist(list):
    newlist = []
    for i in range(len(list)):
        if (list[i] <> ""):
            newlist.append(list[i])
    return newlist

# Get the difference of two lists
def listdiff(list1, list2):
    difflist = []
    for i in range(len(list1)):
        if (not list1[i] in list2):
            difflist.append(list1[i])
    for i in range(len(list2)):
        if (not list2[i] in list1):
            difflist.append(list2[i])
    return difflist


"""

Time functions

"""

# Converts time in seconds to a nicely formatted string
def nicetime(seconds):
    minutes, seconds = divmod(seconds, 60)
    return "%s:%s" % (zfill(minutes, 2), zfill(seconds, 2))


"""

File and directory operations

"""

# Creates a directory listing, optional including full filenames and recursive
def ls(directory, recursive = FALSE, includepath = FALSE, onlydirectories = FALSE):
    longlist = []
    allfiles = os.listdir(directory)
    if (directory == "."): directory = ""
    for i in range(len(allfiles)):
        filename = "%s%s" % (directory, allfiles[i])
        longfilename = "%s/%s" % (os.path.abspath(directory), filename)
        if os.path.isdir(filename):
            if (includepath == TRUE):
                longlist.append(longfilename)
            else:
                longlist.append(filename)
            if (recursive == TRUE):
                subdirlist = ls("%s/" % filename, recursive, includepath, onlydirectories)
                if (len(subdirlist) > 0):
                    longlist = longlist + subdirlist
        else:
            if (onlydirectories == FALSE):
                if (includepath == TRUE):
                    longlist.append(longfilename)
                else:
                    longlist.append(filename)
    return longlist

# Remove a file or directory (recursive) and return the number of directories, files, bytes & links deleted
def rm(file, maximum = -1, deleted = 0):
    directories = files = bytes = links = 0L
    if os.path.islink(file):
        if (maximum == -1) or ((maximum != -1) and (deleted < maximum)):
            links = links + 1
            unlink(file)
    elif os.path.isfile(file):
        filesize = lstat(file)[6]
        if (maximum == -1) or ((maximum != -1) and ((deleted + filesize) < maximum)):
            bytes = bytes + filesize
            files = files + 1
            unlink(file)
    else:
        filelist = os.listdir(file)
        for i in range(len(filelist)):
            result = rm(file + "/" + filelist[i], maximum, bytes)
            directories = directories + result[0]
            files = files + result[1]
            bytes = bytes + result[2]
            links = links + result[3]
        filelist = listdir(file)
        if (maximum == -1) or ((len(filelist) == 0) and (maximum != -1) and (deleted < maximum)):
            rmdir(file)
            directories = directories + 1
    return [directories, files, bytes, links]

# Creates a whole direcotry tree
def mkdirtree(directory):
    if (directory[-1:] == "/"):
        directory = directory[:-1]
    if (not os.path.exists(directory)):
        pwd = getcwd()
        splitted = split(directory, "/")
        path = "/"
        for i in range(len(splitted)):
            path = "%s/%s" % (path, splitted[i])
            if not os.path.exists(path):
                mkdir(path)
        chdir(pwd)

# Calculate the disk usage of a given directory
def du(file):
    if (os.path.islink(file)) or (os.path.isfile(file)):
        size = lstat(file)[6]
    else:
        filelist = listdir(file)
        size = 0
        for i in range(len(filelist)):
            size = size + du(file + "/" + filelist[i])
    return size

# Check the path to find some specified program
def which(file):
    if file == "":
        return ""
    for item in split(os.environ["PATH"], os.pathsep):
        filename="%s%s%s" % (item, os.sep, file)
        if os.path.exists(filename):
            return filename
    return ""

# Copy a source file to a target file
def filecopy(source, target):
    sourcefile = open(source, "r")
    targetfile = open(target, "w")
    targetfile.write(sourcefile.read())
    targetfile.close()
    sourcefile.close()

# Include a file into a already open file
def includefile(source, target):
    sourcefile = open(source, "r")
    line = "line"
    while (line != ""):
        line = sourcefile.readline()
        if (line != ""):
            target.write(line)
    sourcefile.close()

def listdir(path, incfiles = TRUE, incdirs = TRUE, inclinks = TRUE):
    files = os.listdir(path)
    if (incfiles == TRUE) and (incdirs == TRUE) and (inclinks == TRUE):
        cleanfiles = files
    else:
        cleanfiles = []
        if path[-1] != "/":
            path = "%s/" % path
        for i in range(len(files)):
            file = "%s%s" % (path, files[i])
            addFile = TRUE
            if (incfiles == FALSE) and (os.path.isfile(file)):
                addFile = FALSE
            if (incdirs == FALSE) and (os.path.isdir(file)):
                addFile = FALSE
            if (inclinks == FALSE) and (os.path.islink(file)):
                addFile = FALSE
            if (addFile == TRUE):
                cleanfiles.append(files[i])
    cleanfiles.sort()
    return cleanfiles

# Implements the functionality of the unix command "touch"
def touch(filename):
    if (os.path.exists(filename)):
        now = time()
        utime(filename, (now, now))
    else:
        file = open(filename, "w")
        file.close()

"""

Pipe functions

"""

def _readlines(opendescriptor):
    line = None
    output = []
    while (line != ""):
        line = opendescriptor.readline()
        if (line != ""):
            output.append(line)
    return output
        
# Read the complete output of a command
def cmdoutput(command, strip = FALSE, waitpid = FALSE):
    pipe = popen3(command)
    error = pipe[2].readline()
    sys.stdout.write(error)

    # Note: this is a workaround, since pipe.readlines() doesn't work with python 2.x and
    # gnome-python (it throws an IOError: (0, 'Error')), so check this back somewhen !
    # output = pipe[0].readlines()
    
    output = _readlines(pipe[0])
    if (strip == TRUE):
        output = striplist(output)
    pipe[0].close()
    pipe[1].close()
    pipe[2].close()
    del pipe
    pipe = None
    if (waitpid == TRUE):
        os.wait()
    return output

"""

General functions

"""

# Handler, which waits for children to be closed
def SigChldHandler(signal_number, stack_frame):
    # Reinstall the sigchild handler, because it gets deleted on libc5/sysv machines
    signal.signal(signal.SIGCHLD, SigChldHandler)
    pid = -1
    while (pid != 0):
        try:
            pid = os.waitpid(0, os.WNOHANG)[0]
        except:
            pid = 0

# ANSI text (fore- & background, bold font)
def ansi(text, fg = WHITE, bg = BLACK, bold = FALSE):
    if (bold == TRUE):
        fg = "%s;1" % fg
    bg = str(atoi(bg) + 10)
    text = "\033[%s;%sm%s\033[0m" % (bg, fg, text)
    return text


"""

Extended version of urlparse (includes username & password)

"""

# extended urlparse function (including username, password)
# returns: <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
# returns: <scheme>://<user>:<password>@<netloc>:<port>/<path>;<params>?<query>#<fragment>

urlparseorg = urlparse
_services_cache = {}

def readservices():
    global _services_cache
    services = {}
    if (len(_services_cache) == 0):
        if (os.path.exists("/etc/services")):
            file = open("/etc/services", "r")
            lines = file.readlines()
            file.close()
            for i in range(len(lines)):
                line = strip(lines[i])
                if (find(line, "#") != -1):
                    line = strip(line[:find(line, "#")])
                if (len(line) > 0):
                    nameportpair = split(line)
                    if (len(nameportpair) > 1):
                        name, port = nameportpair[:2]
                        if find(port, "/") != -1:
                            port, protocol = split(port, "/")
                            services[name] = port
        else:
            services["ftp"] = 21
            services["http"] = 80
        _services_cache = copy.copy(services)
    else:
        services = copy.copy(_services_cache)
    return services

def urlparse(url):
    scheme, netloc, path, params, query, fragment = urlparseorg(url)
    services = readservices()
    user = password = ""
    if services.has_key(scheme):
        port = atoi(services[scheme])
    else:
        port = 0
    if (find(netloc, "@") != -1):
        user, netloc = split(netloc, "@")
    if (find(user, ":") != -1):
        user, password = split(user, ":")
    if (find(netloc, ":") != -1):
        netloc, port = split(netloc, ":")
        port = atoi(port)
    if (path == ""):
        path = "/"
    tuple = scheme, user, password, netloc, port, path, params, query, fragment
    return tuple



"""

Python logging module - Version 0.3

Loglevels: 
    
    LOGLEVEL_NONE, LOGLEVEL_NORMAL, LOGLEVEL_VERBOSE, LOGLEVEL_DEBUG

Format-Parameters:
    
    %C ..... class-name
    %D ..... program duration
    %d ..... duration for the last step (last output)
    %F ..... function name
    %L ..... logtype (Error, Warning, ...)
    %M ..... message
    %N ..... Line-number
    %T ..... current time

Pre-defined Formats:

    FMT_SHORT, FMT_MEDIUM, FMT_LONG, FMT_DEBUG

"""

# Logging levels

LOGLEVEL_NONE = 1 << 0
LOGLEVEL_NORMAL = 1 << 1
LOGLEVEL_VERBOSE = 1 << 2
LOGLEVEL_DEBUG = 1 << 3

# Pre-defined format strings

FMT_SHORT = "%M"
FMT_MEDIUM = "[ %C.%F ] %D: %M"
FMT_LONG = "%T %L %C [%F] - %M"
FMT_DEBUG = "%T [%D (%d)] %L %C [%F (%N)] - %M"

# Configuration files

CONFIGURATION_FILES = {}
CONFIGURATION_FILES[1] = "log4py.conf"                    # local directory
CONFIGURATION_FILES[2] = "$HOME/.log4py.conf"             # hidden file in the home directory
CONFIGURATION_FILES[3] = "/etc/log4py.conf"               # system wide file

# The following constants are of internal interest only

# Message constants (used for ansi colors and for logtype %L)

MSG_DEBUG = 1 << 0
MSG_WARN = 1 << 1
MSG_ERROR = 1 << 2
MSG_INFO = 1 << 3

# Boolean constants

TRUE = "TRUE"
FALSE = "FALSE"

# Color constants

BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
PURPLE = 35
AQUA = 36
WHITE = 37

LOG_MSG = {MSG_DEBUG: "DEBUG", MSG_WARN: "WARNING", MSG_ERROR: "ERROR", MSG_INFO: "INFO"}
LOG_COLORS = {MSG_DEBUG: [WHITE, BLACK, FALSE], MSG_WARN: [WHITE, BLACK, FALSE], MSG_ERROR: [WHITE, BLACK, TRUE], MSG_INFO: [WHITE, BLACK, FALSE]}
LOG_LEVELS = { "DEBUG": LOGLEVEL_DEBUG, "VERBOSE": LOGLEVEL_VERBOSE, "NORMAL": LOGLEVEL_NORMAL, "NONE": LOGLEVEL_NONE }

from time import time, strftime, localtime
from types import StringType, ClassType, InstanceType
from string import zfill, atoi, lower, upper
from re import sub
from ConfigParser import ConfigParser

import sys
import traceback
import os

# This is the main class for the logging module

class Category:

    # class initalization & customization - don't use sys.stdout as default target - copy.deepcopy won't work otherwise

    def __init__(self, classid = None, useconfigfiles = TRUE, customconfigfiles = []):

        self.__Category_setdefaults()
        if (useconfigfiles == TRUE):
            self.__Category_appendconfigfiles(customconfigfiles)
            self.__Category_loadconfigfile()

        self.__Category_timeinit = time()
        self.__Category_timelaststep = self.__Category_timeinit
        if (classid == None):
            self.__Category_classname = "Main"
        else:
            if (type(classid) == StringType):
                self.__Category_classname = classid
            elif (type(classid) == ClassType):
                self.__Category_classname = classid.__name__
            elif (type(classid) == InstanceType):
                self.__Category_classname = classid.__class__.__name__

    # Methods to set properties

    def set_target(self, target):
        self.__Category_target = target

    def set_loglevel(self, loglevel):
        self.__Category_loglevel = loglevel

    def set_formatstring(self, formatstring):
        self.__Category_formatstring = formatstring

    def set_use_ansi_codes(self, useansicodes):
        self.__Category_useansicodes = useansicodes

    def set_time_format(self, timeformat):
        self.__Category_timeformat = timeformat

    # Method to get properties

    def get_loglevel(self):
        return self.__Category_loglevel

    def get_formatstring(self):
        return self.__Category_formatstring

    def get_use_ansi_codes(self):
        return self.__Category_useansicodes

    def get_time_format(self):
        return self.__Category_timeformat

    # Methods to actually print messages

    def debug(self, message):
        if (self.__Category_loglevel >= LOGLEVEL_DEBUG):
            self.__Category_showmessage(message, MSG_DEBUG)

    def warn(self, message):
        if (self.__Category_loglevel >= LOGLEVEL_VERBOSE):
            self.__Category_showmessage(message, MSG_WARN)

    def error(self, message):
        if (self.__Category_loglevel >= LOGLEVEL_NORMAL):
            self.__Category_showmessage(message, MSG_ERROR)

    def info(self, message):
        if (self.__Category_loglevel >= LOGLEVEL_NORMAL):
            self.__Category_showmessage(message, MSG_INFO)

    # Private method of the Category class - you never have to use those directly

    def __Category_tracestack(self):
        stack = traceback.extract_stack()
        self.__Category_linenumber = stack[-4][1]
        self.__Category_functionname = stack[-4][2]
        if (self.__Category_functionname == "?"):
            self.__Category_functionname = "Main"

    def __Category_setdefaults(self):
        self.__Category_target = None                         # None = sys.stdout
        self.__Category_formatstring = FMT_LONG
        self.__Category_loglevel = LOGLEVEL_NORMAL
        self.__Category_useansicodes = FALSE
        self.__Category_functionname = ""
        self.__Category_linenumber = -1
        self.__Category_timeformat = "%d.%m.%Y %H:%M:%S"

    def __Category_loadconfigfile(self):
        priorities = CONFIGURATION_FILES.keys()
        priorities.sort()
        configfilename = ""
        for i in range(len(priorities)):
            filename = CONFIGURATION_FILES[priorities[i]]
            filename = sub("\$HOME", os.environ["HOME"], filename)
            if (os.path.exists(filename)):
                configfilename = filename
                break
        if (configfilename != ""):
            parser = ConfigParser()
            parser.read(configfilename)
            section = "log4py"
            for i in range(len(parser.options(section))):
                option = lower(parser.options(section)[i])
                value = parser.get(section, option)
                if (option == "format"):
                    self.set_formatstring(value)
                elif (option == "loglevel"):
                    self.set_loglevel(LOG_LEVELS[upper(value)])
                elif (option == "timeformat"):
                    self.set_time_format(value)
                elif (option == "ansi"):
                    self.__Category_useansicodes = upper(value)
                elif (option == "target"):
                    if (lower(value) == "stdout"):
                        self.set_target(None)
                    elif (lower(value) == "stderr"):
                        self.set_target(sys.stderr)
                    else:
                        self.set_target(value)
            return TRUE
        else:
            return FALSE

    def __Category_appendconfigfiles(self, filenames):
        filenames.reverse()
        for i in range(len(filenames)):
            keys = CONFIGURATION_FILES.keys()
            CONFIGURATION_FILES[min(keys) - 1] = filenames[i]

    def __Category_showmessage(self, message, messagesource):
        currenttime = time()
        self.__Category_tracestack()
        timedifference = "%.3f" % (currenttime - self.__Category_timeinit)
        timedifflaststep = "%.3f" % (currenttime - self.__Category_timelaststep)
        self.__Category_timelaststep = currenttime
        milliseconds = int(round((currenttime - long(currenttime)) * 1000))
        timeformat = sub("%S", "%S." + (zfill(milliseconds, 3)), self.__Category_timeformat)
        currenttime = strftime(timeformat, localtime(currenttime))

        line = self.__Category_formatstring
        line = sub("%C", self.__Category_classname, line)
        line = sub("%D", timedifference, line)
        line = sub("%d", timedifflaststep, line)
        line = sub("%F", self.__Category_functionname, line)
        if (self.__Category_useansicodes == TRUE):
            line = sub("%L", self.__Category_ansi(LOG_MSG[messagesource], messagesource), line)
            line = sub("%M", self.__Category_ansi(message, messagesource), line)
        else:
            line = sub("%L", LOG_MSG[messagesource], line)
            line = sub("%M", message, line)
        line = sub("%N", str(self.__Category_linenumber), line)
        line = sub("%T", currenttime, line)

        if (self.__Category_target == sys.stdout) or (self.__Category_target == sys.stderr):
            self.__Category_target.write("%s\n" % line)
        elif (self.__Category_target == None):
            sys.stdout.write("%s\n" % line)
        else:
            file = open(self.__Category_target, "a")
            file.write("%s\n" % line)
            file.close()

    # Convert text to text including ansi codes

    def __Category_ansi(self, text, messagesource):
        bold = LOG_COLORS[messagesource][2]
        fg = str(LOG_COLORS[messagesource][0])
        bg = LOG_COLORS[messagesource][1]
        if (bold == TRUE):
            fg = "%s;1" % fg
        bg = bg + 10
        text = "\033[%d;%sm%s\033[0m" % (bg, fg, text)
        return text
